
#pragma once

#include <root/component.h>
#include <tee/session_component_base.h>
#include <tee/session_env.h>

namespace Tee {
	using Genode::Allocator;
	using Genode::Arg_string;
	using Genode::Env;
	using Genode::Insufficient_ram_quota;
	using Genode::error;
	using Genode::max;
	using Genode::size_t;

	template <class SESSION_COMPONENT> class Root;
};


template <class SESSION_COMPONENT>
class Tee::Root : public Genode::Root_component<SESSION_COMPONENT,
                                                Genode::Single_client>
{
	private:

		Env       &_env;
		Allocator &_md_alloc;

	protected:

		SESSION_COMPONENT *_create_session(const char *args) override
		{
			const Session_label label        { label_from_args(args) };
			const auto          ram_quota    { parse_ram_quota(args) };
			const auto          cap_quota    { parse_cap_quota(args) };
			const auto          tx_buf_size  { parse_tx_buffer_size(args) };
			const auto          rx_buf_size  { parse_rx_buffer_size(args) };

			try {

				// create session environment temporarily on the stack
				Session_env session_env_stack {  _env, Ram_quota { ram_quota }, Cap_quota { cap_quota } };

				// alloc/attach RAM block
				Ram_dataspace_capability ram_ds { session_env_stack.alloc(sizeof(Session_env) +
				                                                          sizeof(SESSION_COMPONENT), CACHED) };

				try {
					// move session env to base of the block
					void*         ram_ptr     { session_env_stack.attach(ram_ds) };
					Session_env  &session_env { *construct_at<Session_env>(ram_ptr, session_env_stack) };

					try {
						log("incoming session label='", label, "'");
						auto session { construct_at<SESSION_COMPONENT>((void*)((addr_t)ram_ptr + sizeof(Session_env)),
						                                               session_env,
						                                               tx_buf_size,
						                                               rx_buf_size,
						                                               ram_ds,
						                                               label) };
						return session;
					} catch (Out_of_ram) {
						error("not enough RAM to create session for ", label);
						Session_env session_env_stack { session_env };
						session_env_stack.detach(ram_ptr);
						session_env_stack.free(ram_ds);
						throw Insufficient_ram_quota();
					} catch (Out_of_caps) {
						error("not enough CAPS to create session for ", label);
						Session_env session_env_stack { session_env };
						session_env_stack.detach(ram_ptr);
						session_env_stack.free(ram_ds);
						throw Insufficient_cap_quota();
					}
				} catch (Out_of_ram) {
					error("not enough RAM to create session session_env for ", label);
					session_env_stack.free(ram_ds);
					throw Insufficient_ram_quota();
				} catch (Out_of_caps) {
					error("not enough CAPS to create session session_env for ", label);
					session_env_stack.free(ram_ds);
					throw Insufficient_cap_quota();
				}
			} catch (Out_of_ram) {
				error("not enough RAM ", label);
				throw Insufficient_ram_quota();
			} catch (Out_of_caps) {
				error("not enough CAPS ", label);
				throw Insufficient_cap_quota();
			} catch (...) {
				error("session creation failed for ", label);
				throw Service_denied();
			}
		}

	public:

		Root(Genode::Env &env, Genode::Allocator &md_alloc)
		: Genode::Root_component<SESSION_COMPONENT, Genode::Single_client>(&env.ep().rpc_ep(), &md_alloc),
			_env(env), _md_alloc(md_alloc)
		{ }
};

